from ipaddress import IPv4Address, IPv6Address
import socket
from threading import Event, Thread
from typing import List, Optional, Tuple, Union
from wgstarman.cli.protocol import ErrorCode, read_message, send_message
from wgstarman.cli.protocol.messages import ErrorResponse, Message, ResolveRequest, ResolveResponse
from wgstarman.cli.protocol.utils import MessageEncDec

from wgstarman.wg_conf import WireGuardConf


class ResolverServer:
    bind_address: Union[IPv4Address, IPv6Address]
    bind_port: int
    config: WireGuardConf
    end_signal: Event

    def __init__(self, bind_address: Union[IPv4Address, IPv6Address], bind_port: int, config: WireGuardConf, end_signal: Event):
        self.bind_address = bind_address
        self.config = config
        self.bind_port = bind_port
        self.end_signal = end_signal

    def resolve(self, host_name: str) -> Optional[List[str]]:
        return next((peer.allowed_ips for peer in self.config.peers if peer.name == host_name), None)

    def listen(self):
        family = socket.AF_INET6 if isinstance(self.bind_address, IPv6Address) else socket.AF_INET
        sock = socket.create_server(
            address=(self.bind_address.exploded, self.bind_port), family=family)

        thread = Thread(target=self.listen_routine, args=[sock])
        thread.start()

        # Wait for the thread to end
        thread.join()

    def listen_routine(self, sock: socket.socket) -> None:
        threads: List[Thread] = []
        sock.settimeout(1)

        while not self.end_signal.is_set():
            try:
                sock.listen()
                connection, address = sock.accept()
                thread = Thread(target=self.request_handler, args=[connection, address])
                thread.start()
                threads.append(thread)
            except socket.timeout:
                pass

        sock.close()

        # Wait for all the threads
        for thread in threads:
            if thread.is_alive():
                thread.join()

    def request_handler(self, connection: socket.socket, address: Tuple[str, int]):
        raw_msg = read_message(connection)

        message: ResolveRequest = MessageEncDec.loads(raw_msg)

        if not message.instance_of(ResolveRequest):
            send_message(connection, ErrorResponse(
                ErrorCode.UNHANDLED_REQUEST.value, 'Unhandled request'))

        send_message(connection, ResolveResponse(
            message.host_name, self.resolve(message.host_name)))

        connection.close()



import ipaddress
import logging
import socket
from argparse import (SUPPRESS, ArgumentDefaultsHelpFormatter, ArgumentParser,
                      Namespace, _SubParsersAction)
from typing import Union

from wgstarman.cli.common import CLICommand
from wgstarman.cli.protocol import (AcknowledgeResponse, ErrorResponse,
                                    IPAddressHoldRequest, IPAddressRequest,
                                    IPAddressResponse, MessageEncDec, ProtocolException,
                                    read_message, send_message)
from wgstarman.cli.wgstarman_resolvers import WGStarManResolver
from wgstarman.wg_conf import Interface, Peer, WireGuardConf
from wgstarman.wgcli.exec import WireGuardCLI

LOG_FORMAT = '%(levelname)9s %(message)s'


class PeerCommand(CLICommand):
    task_name = 'peer'

    @staticmethod
    def configure_logging(args: Namespace) -> None:
        logging.basicConfig(format=LOG_FORMAT, level=logging.DEBUG if args.debug else logging.INFO)

    def decorate_subparsers(self, subparsers: _SubParsersAction) -> None:
        parser: ArgumentParser = subparsers.add_parser(
            self.task_name, formatter_class=ArgumentDefaultsHelpFormatter)
        required = parser.add_argument_group('required')
        optional = parser.add_argument_group('optional')

        required.add_argument(
            '--server-address', help="the server's IPv4 or IPv6 address", required=True, default=SUPPRESS)
        required.add_argument(
            '--device-name', help='the name of both the network device and the configuration file', required=True, default=SUPPRESS)
        required.add_argument(
            '--psk', help='the pre-shared key, given by the server at startup', required=True, default=SUPPRESS)

        optional.add_argument('--server-port', default=1194, type=int,
                              help="the server's port", required=False)
        optional.add_argument('--keep-alive', default=False, action='store_true',
                              help='keep the connection to the server alive (to avoid connection drop in case of NAT)', required=False)
        optional.add_argument('--host-name',
                              help='name of the peer, resolvable via wgstarman resolv', required=False)
        optional.add_argument('--overwrite', default=False, action='store_true',
                              help="overwrite if a configuration already exists", required=False)
        optional.add_argument('--debug', default=False, action='store_true',
                              help='enable debug log', required=False)

    def can_handle(self, task_name: str) -> bool:
        return self.task_name == task_name

    def handle(self, args: Namespace) -> None:
        encdec_key = args.psk.encode('utf8')

        try:
            ipaddress.ip_address(args.server_address)
            server_address: str = args.server_address
        except:
            server_address = socket.gethostbyname(args.server_address)

        try:
            conf = WireGuardConf.parse(args.device_name)
        except PermissionError:
            logging.error('wgstarman re')
            return

        if conf is not None:
            if not args.overwrite:
                logging.error(
                    f'Configuration file for device {args.device_name} already exists. Use --overwrite to continue.')

                return
            else:
                logging.info('Configuration detected. Overwriting.')

        private_key = conf.interface.private_key if conf is not None else WireGuardCLI.gen_private_key()
        public_key = WireGuardCLI.get_public_key(private_key)

        logging.debug(f'Connecting to central peer ({server_address} : {args.server_port})')
        sock = socket.create_connection((server_address, args.server_port), 5)

        try:
            # Request an IP address assigned to the public key
            send_message(sock, IPAddressRequest(public_key), encdec_key)
            response: Union[IPAddressResponse, ErrorResponse] = MessageEncDec.loads(
                read_message(sock, encdec_key))
            if response.instance_of(ErrorResponse):
                logging.error(f'{response.error_code}: {response.message}')

                return

            conf = WireGuardConf(
                Interface(private_key=private_key, address=[
                    response.peer_address]),
                [Peer(
                    public_key=response.server_public_key, allowed_ips=[response.peer_allowed_ips],
                    endpoint=f'{server_address}:{args.server_port}', persistend_keep_alive=25 if args.keep_alive else None)]
            )
            resolv_conf = WGStarManResolver.load(args.device_name)
            resolv_conf.resolver_addresses = response.resolver_addresses

            # Send the IP address hold message
            sock = socket.create_connection((server_address, args.server_port), 5000)
            send_message(sock, IPAddressHoldRequest(
                conf.interface.address[0], args.host_name), encdec_key)
            response: Union[AcknowledgeResponse, ErrorResponse] = MessageEncDec.loads(
                read_message(sock, encdec_key))
            if response.instance_of(ErrorResponse):
                logging.error(f'{response.error_code}: {response.message}')
                return

            conf.save(args.device_name)
            resolv_conf.save()

            WireGuardCLI.up(args.device_name)
        except ProtocolException as e:
            logging.error(e)



from argparse import SUPPRESS, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
import logging
import os
import socket
from typing import Union
from wgstarman.cli.common import CLICommand
from wgstarman.cli.peer import PeerCommand
from argparse import _SubParsersAction
from wgstarman.cli.protocol import read_message, send_message
from wgstarman.cli.protocol.messages import ErrorResponse, ResolveRequest, ResolveResponse
from wgstarman.cli.protocol.utils import MessageEncDec

from wgstarman.cli.wgstarman_resolvers import WGStarManResolver


class ResolvCommand(CLICommand):
    task_name = 'resolv'

    @staticmethod
    def configure_logging(args: Namespace) -> None:
        return PeerCommand.configure_logging(args)

    @staticmethod
    def ensure_wireguard() -> bool:
        return False

    def decorate_subparsers(self, subparsers: _SubParsersAction) -> None:
        parser: ArgumentParser = subparsers.add_parser(
            self.task_name, formatter_class=ArgumentDefaultsHelpFormatter)
        required = parser.add_argument_group('required')
        optional = parser.add_argument_group('optional')

        required.add_argument(
            '--device-name', '-d', help='the device associated with the network', required=True, default=SUPPRESS)
        required.add_argument('host_name')

        optional.add_argument(
            '--resolv-port', '-p', help='the resolv port (server listen port + 1)', type=int, default=1195)
        optional.add_argument('--debug', default=False, action='store_true',
                              help='enable debug log', required=False)

    def can_handle(self, task_name: str) -> bool:
        return self.task_name == task_name

    def handle(self, args: Namespace) -> None:
        resolv_conf = WGStarManResolver.load(args.device_name)
        if not resolv_conf.resolver_addresses:
            logging.error(f'Unknown resolver for device {args.device_name}')

            return

        # special case: resolv will return the resolver's IP addresses
        if args.host_name == 'resolv':
            print('\n'.join(resolv_conf.resolver_addresses))

            return

        host_name = args.host_name
        resolver_port = args.resolv_port
        for resolv_address in resolv_conf.resolver_addresses:
            try:
                sock = socket.create_connection((resolv_address, resolver_port), 1)
                send_message(sock, ResolveRequest(host_name))
                response = read_message(sock)
                message: Union[ResolveResponse, ErrorResponse] = MessageEncDec.loads(response)

                if message.instance_of(ErrorResponse):
                    logging.error(f'{message.error_code}: {message.message}')

                    os._exit(1)

                print('\n'.join(message.ip_addresses))

                return
            except socket.timeout:
                pass

        logging.error('Unable to contact resolv server')
        os._exit(2)



import logging
import subprocess
from functools import reduce


class WireGuardCLI:
    logger = logging.getLogger('WireGuard executor')

    @staticmethod
    def ensure_installation() -> bool:
        wg_check = subprocess.run('which wg', shell=True,
                                  stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)
        wg_quick_check = subprocess.run(
            'which wg-quick', shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)

        if wg_check.stderr or wg_quick_check.stderr:
            WireGuardCLI.logger.error(
                'WireGuard not found. '
                + 'Ensure that you have both `wg` and `wg-quick` commands installed.')

            return False

        wg_permissions_check = subprocess.run(
            'wg showconf whatever', shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)

        if wg_permissions_check.stderr and wg_permissions_check.stderr.decode('utf8').strip() \
                .endswith('Operation not permitted'):
            WireGuardCLI.logger.error(
                "Can't run WireGuard commands with the current privileges. "
                + "Ensure you've running the command with administrative privileges.")

            return False

        return True

    @staticmethod
    def gen_private_key() -> str:
        result = subprocess.run('wg genkey', shell=True, stdout=subprocess.PIPE)

        return result.stdout.decode('utf8').strip()

    @staticmethod
    def gen_preshared_key() -> str:
        result = subprocess.run('wg genpsk', shell=True, stdout=subprocess.PIPE)

        return result.stdout.decode('utf8').strip()

    @staticmethod
    def get_public_key(private_key: str) -> str:
        result = subprocess.run(f'wg pubkey', shell=True,
                                stdout=subprocess.PIPE, input=private_key.encode('utf8'))

        return result.stdout.decode('utf8').strip()

    @staticmethod
    def up(device_name) -> bool:
        subprocess.run(f'wg-quick down {device_name}', shell=True, stderr=subprocess.DEVNULL)

        result = subprocess.run(f'wg-quick up {device_name}', shell=True)

        if result.returncode:
            WireGuardCLI.logger.error(f'Unable to up device {device_name}')

            return False

        return True

    @staticmethod
    def down(device_name) -> bool:
        result = subprocess.run(f'wg-quick down {device_name}',
                                shell=True, stderr=subprocess.DEVNULL)

        if result.returncode:
            WireGuardCLI.logger.error(f'Unable to down device {device_name}')

            return False

        return True

    def hot_reload(device_name) -> bool:
        results = [
            subprocess.run(f'wg-quick down {device_name}', shell=True),
            subprocess.run(f'wg-quick up {device_name}', shell=True),
        ]

        exit_code = reduce(lambda a, b: a + b, map(lambda a: a.returncode, results))
        if exit_code:
            WireGuardCLI.logger.error(f'Unable to hot reload device {device_name}')

            return False

        return True



from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

from wgstarman.cli.common import DEFAULT_WGSTARMAN_CONF_PATH, ETC_DIR_MODE, WG_CONF_FILE_MODE


@dataclass
class PresharedKey:
    public_key: str
    preshared_key: str

    def parse(line: str) -> 'PresharedKey':
        public_key, preshared_key = line.split(' = ')

        return PresharedKey(public_key.strip(), preshared_key.strip())

    def __str__(self) -> str:
        return f'{self.public_key} = {self.preshared_key}'


@dataclass
class WGStarManConf:
    preshared_keys: List[PresharedKey]

    @staticmethod
    def load(conf_path: str = DEFAULT_WGSTARMAN_CONF_PATH) -> 'WGStarManConf':
        path = Path(conf_path)

        if not path.exists():
            return WGStarManConf([])

        preshared_keys = []
        for line in path.read_text(encoding='utf8').split('\n'):
            if line:
                preshared_keys.append(PresharedKey.parse(line))

        return WGStarManConf(preshared_keys)

    def save(self, conf_path: str = DEFAULT_WGSTARMAN_CONF_PATH) -> None:
        path = Path(conf_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.parent.chmod(ETC_DIR_MODE)

        path.write_text('\n'.join(map(lambda psk: str(psk), self.preshared_keys)) + '\n')
        path.chmod(WG_CONF_FILE_MODE)

    def get_preshared_key(self, public_key: str) -> Optional[str]:
        return next((psk.preshared_key for psk in self.preshared_keys if psk.public_key == public_key), None)

    def set_preshared_key(self, public_key: str, preshared_key: str) -> None:
        old_psk = next((psk for psk in self.preshared_keys if psk.public_key == public_key), None)
        if old_psk:
            self.preshared_keys.remove(old_psk)

        self.preshared_keys.append(PresharedKey(public_key, preshared_key))

        self.save()



from dataclasses import dataclass, field
from typing import ClassVar, List, Optional, Type, TypeVar


class Message:
    message_type: str = 'UnknownMessage'

    def instance_of(self, cls: Type['Message']) -> bool:
        return type(self) == cls


@dataclass
class IPAddressRequest(Message):
    public_key: str
    peer_name: Optional[str] = field(default=None)

    message_type: ClassVar[str] = 'request_ip_address'


@dataclass
class IPAddressHoldRequest(Message):
    ip_address: str
    host_name: Optional[str]

    message_type: ClassVar[str] = 'request_ip_address_hold'


@dataclass
class IPAddressResponse(Message):
    server_public_key: str
    peer_address: str
    peer_allowed_ips: str
    resolver_addresses: List[str]

    message_type: ClassVar[str] = 'ip_address_response'


@dataclass
class ResolveRequest(Message):
    host_name: str

    message_type: ClassVar[str] = 'resolv_request'


@dataclass
class ResolveResponse(Message):
    host_name: str
    ip_addresses: List[str]

    message_type: ClassVar[str] = 'resolv_response'


@dataclass
class ErrorResponse(Message):
    error_code: int
    message: str

    message_type: ClassVar[str] = 'error_response'


class AcknowledgeResponse(Message):
    message_type: ClassVar[str] = 'ack'


T = TypeVar('T', bound=Message)
REGISTERED_MESSAGES: List[Type[T]] = [
    IPAddressRequest, IPAddressHoldRequest, IPAddressResponse,
    ResolveRequest, ResolveResponse,
    ErrorResponse, AcknowledgeResponse
]



from threading import Lock
from typing import Callable


def syncronized(lock: Lock) -> Callable:
    def synchronized_decorator(fn: Callable) -> Callable:
        def _synchronized_decorator(*args, **kwargs):
            with lock:
                return fn(*args, **kwargs)

        return _synchronized_decorator

    return synchronized_decorator



import logging
import socket
from enum import Enum
from typing import Any, Optional

from cryptography.fernet import Fernet
from wgstarman.cli.protocol.messages import (
    AcknowledgeResponse, ErrorResponse, IPAddressHoldRequest, IPAddressRequest, IPAddressResponse, Message
)
from wgstarman.cli.protocol.utils import MessageEncDec

__all__ = [
    # messages
    IPAddressRequest, IPAddressResponse, IPAddressHoldRequest, AcknowledgeResponse, ErrorResponse, Message
]


class ProtocolException(Exception):
    pass


class ErrorCode(Enum):
    NETWORK_IS_FULL = 10001,
    INVALID_IP_ADDRESS = 10002,
    HOST_NAME_THEFT = 10003,
    UNABLE_TO_RELOAD_CONFIGURATION = 10004,
    UNHANDLED_REQUEST = 10005,


INVALID_TOKEN_LEN = 0xFFFF


def send_message(conn: socket.socket, msg: Any, encrypt_key: Optional[bytes] = None) -> None:
    msg = MessageEncDec.dumps(msg)
    logging.debug(f'Sending {msg}')

    if encrypt_key:
        fernet = Fernet(encrypt_key)
        msg = fernet.encrypt(msg.encode('utf8'))
    else:
        msg = msg.encode('utf8')
    msg = len(msg).to_bytes(4, 'big', signed=False) + msg

    conn.send(msg)


def send_invalid_token_message(conn: socket.socket) -> None:
    msg = INVALID_TOKEN_LEN.to_bytes(4, 'big', signed=False)

    conn.send(msg)


def read_message(conn: socket.socket, decrypt_key: Optional[bytes] = None) -> str:
    msg_len = int.from_bytes(conn.recv(4), 'big', signed=False)
    if msg_len == INVALID_TOKEN_LEN:
        raise ProtocolException('The pre-shared key is invalid.')

    msg = conn.recv(msg_len)

    if decrypt_key:
        fernet = Fernet(decrypt_key)
        msg = fernet.decrypt(msg).decode('utf8')
    else:
        msg = msg.decode('utf8')

    logging.debug(f'Receiving {msg}')

    return msg



from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

from wgstarman.cli.common import DEFAULT_WGSTARMAN_ETC_PATH, RESOLVERS_ETC_DIR_MODE, RESOLVERS_FILE_MODE


@dataclass
class WGStarManResolver:
    device_name: str
    resolver_addresses: List[str]

    @staticmethod
    def get_conf_path(device_name: str, conf_path: str = DEFAULT_WGSTARMAN_ETC_PATH) -> Path:
        return Path(conf_path).joinpath(f'{device_name}.resolver.conf')

    @staticmethod
    def load(device_name: str, conf_path: str = DEFAULT_WGSTARMAN_ETC_PATH) -> 'WGStarManResolver':
        path = WGStarManResolver.get_conf_path(device_name, conf_path)

        if not path.exists():
            return WGStarManResolver(device_name, [])

        return WGStarManResolver(device_name, list(map(str.strip, path.read_text().split(','))))

    def save(self, conf_path: str = DEFAULT_WGSTARMAN_ETC_PATH) -> 'WGStarManResolver':
        path = WGStarManResolver.get_conf_path(self.device_name, conf_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.parent.chmod(RESOLVERS_ETC_DIR_MODE)

        path.write_text(','.join(self.resolver_addresses))
        path.chmod(RESOLVERS_FILE_MODE)

        return self



import inspect
import json
import logging
from typing import Any, Dict, Optional, Type

from wgstarman.cli.protocol.messages import REGISTERED_MESSAGES, Message, T


class MessageEncDec:
    @staticmethod
    def dumps(obj: Message) -> Optional[str]:
        try:
            dictionary = {'message_type': obj.message_type}
            dictionary.update(vars(obj))

            return json.dumps(dictionary)
        except Exception as e:
            logging.getLogger(__class__.__name__).error(e)

            return None

    @staticmethod
    def loads(raw: str) -> Optional[Message]:
        try:
            jsonObj = json.loads(raw)
            message_type = jsonObj['message_type']
            for message_class in REGISTERED_MESSAGES:
                if message_class.message_type == message_type:
                    return MessageEncDec.from_raw(message_class, jsonObj)
        except Exception as e:
            logging.getLogger(__class__.__name__).error(e)

            return None

    @staticmethod
    def from_raw(cls: Type[T], msg: Dict[str, Any]) -> Optional[T]:
        sig = inspect.signature(cls.__init__)

        try:
            kwargs = {key: value for key, value in msg.items() if key in sig.parameters}

            return cls(**kwargs)

        except Exception as e:
            logging.getLogger(__class__.__name__).error(e)

            return None



import logging
from collections import namedtuple
from dataclasses import dataclass, field
from functools import reduce
from pathlib import Path
from typing import Any, List, Optional, Tuple

from wgstarman.cli.common import ETC_DIR_MODE, WG_CONF_FILE_MODE, WG_DIR_MODE

DEFAULT_WIREGUARD_ETC_DIR = '/etc/wireguard'

KwargResolve = namedtuple('KwargResolver', ['wg_name', 'conf_name', 'type'])


def resolve(resolver: List[KwargResolve], setting: str, value: str) -> Tuple[str, Any]:
    resolve = next((r for r in resolver if r.wg_name == setting), None)

    if resolve:
        if resolve.type == List[str]:
            value = value.strip().split(',')
        elif resolve.type == List[int]:
            value = [int(v) for v in value.strip().split(',')]
        else:
            value = resolve.type(value.strip())

        return (resolve.conf_name, value)

    return None


@dataclass
class Interface:
    private_key: str
    address: List[str]
    listen_port: Optional[int] = field(default=None)
    fw_mark: Optional[int] = field(default=None)

    @staticmethod
    def parse(section: str):
        kwarg_resolver = [
            KwargResolve('PrivateKey', 'private_key', str),
            KwargResolve('Address', 'address', List[str]),
            KwargResolve('ListenPort', 'listen_port', int),
            KwargResolve('FwMark', 'fw_mark', int),
        ]

        kwargs = {}
        for line in section.split('\n')[1:]:
            setting, value = map(str.strip, line.split('=', 1))
            resolved = resolve(kwarg_resolver, setting, value)
            if resolved:
                kwargs[resolved[0]] = resolved[1]

        return Interface(**kwargs)

    def __str__(self) -> str:
        result = []
        result.append('[Interface]')
        result.append(f'PrivateKey = {self.private_key}')
        result.append(f'Address = {",".join(self.address)}')
        if self.listen_port:
            result.append(f'ListenPort = {self.listen_port}')
        if self.fw_mark:
            result.append(f'FwMark = {self.fw_mark}')

        return '\n'.join(result)


@dataclass
class Peer:
    public_key: str
    allowed_ips: List[str]
    preshared_key: Optional[str] = field(default=None)
    endpoint: Optional[str] = field(default=None)
    persistend_keep_alive: Optional[int] = field(default=None)
    name: Optional[str] = field(default=None)

    @staticmethod
    def parse(section: str):
        kwarg_resolver = [
            KwargResolve('PublicKey', 'public_key', str),
            KwargResolve('AllowedIPs', 'allowed_ips', List[str]),
            KwargResolve('PresharedKey', 'preshared_key', str),
            KwargResolve('Endpoint', 'endpoint', str),
            KwargResolve('PersistentKeepalive', 'persistend_keep_alive', int),
            KwargResolve('# Name', 'name', str),
        ]

        kwargs = {}
        for line in section.split('\n')[1:]:
            setting, value = map(str.strip, line.split('=', 1))
            resolved = resolve(kwarg_resolver, setting, value)
            if resolved:
                kwargs[resolved[0]] = resolved[1]

        return Peer(**kwargs)

    def __str__(self) -> str:
        result = []
        result.append('[Peer]')
        if self.name:
            result.append(f'# Name = {self.name}')
        result.append(f'PublicKey = {self.public_key}')
        result.append(f'AllowedIPs = {",".join(self.allowed_ips)}')
        if self.preshared_key:
            result.append(f'PresharedKey = {self.preshared_key}')
        if self.endpoint:
            result.append(f'Endpoint = {self.endpoint}')
        if self.persistend_keep_alive:
            result.append(f'PersistentKeepalive = {self.persistend_keep_alive}')

        return '\n'.join(result)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Peer):
            return False

        def cmp(attr: str) -> bool:
            a = getattr(self, attr)
            b = getattr(other, attr)

            if type(a) == list:
                a = ','.join(a)
                b = ','.join(b)

            return a == b or a is None and b is None

        return reduce(lambda cumu, attr: cumu and cmp(attr), vars(self).keys(), True)


@ dataclass
class WireGuardConf:
    interface: Interface
    peers: List[Peer] = field(default_factory=lambda: [])

    @ staticmethod
    def parse(device_name: str, conf_path: str = DEFAULT_WIREGUARD_ETC_DIR) -> Optional['WireGuardConf']:
        logger = logging.getLogger('WireGuard ConfParser')
        path = Path(conf_path).joinpath(f'{device_name}.conf')

        if not path.exists():
            logger.debug(f'Configuration not found ({path})')

            return None

        sections: List[str] = []
        lines = path.read_text('utf8').split('\n')
        for line in lines:
            line = line.strip()
            if not line:
                continue

            if line.startswith('['):
                sections.append(line)
            else:
                sections[-1] += f'\n{line}'

        interface: Interface = None
        peers: List[Peer] = []
        for section in sections:
            if section.startswith('[Interface]'):
                interface = Interface.parse(section)
            if section.startswith('[Peer]'):
                peers.append(Peer.parse(section))

        if not interface:
            logger.error('Invalid configuration found (no Interface section)')

            return None

        return WireGuardConf(interface, peers)

    def save(self, device_name: str, conf_path: str = DEFAULT_WIREGUARD_ETC_DIR):
        path = Path(conf_path).joinpath(f'{device_name}.conf')
        if not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            path.parent.chmod(WG_DIR_MODE)

        with path.open('w') as fp:
            parts = [str(self.interface)]
            parts.extend([str(peer) for peer in self.peers])
            fp.write('\n\n'.join(parts) + '\n')

        path.chmod(WG_CONF_FILE_MODE)

    def append_peer(self, peer: Peer) -> None:
        old_peer = next((pr for pr in self.peers if peer.public_key == pr.public_key), None)
        if old_peer:
            self.peers.remove(old_peer)

        self.peers.append(peer)

    def find_peer_by_name(self, peer_name: str) -> Optional[Peer]:
        return next((peer for peer in self.peers if peer.name == peer_name), None)



import ipaddress
import logging
import signal
import socket
from argparse import (SUPPRESS, ArgumentDefaultsHelpFormatter, ArgumentParser,
                      Namespace, _SubParsersAction)
from ipaddress import IPv4Address, IPv6Address, IPv6Interface
from threading import Event, Lock, Thread
from typing import Dict, List, Tuple, Union

from cryptography.fernet import Fernet, InvalidToken

from wgstarman.cli.common import CLICommand
from wgstarman.cli.protocol import (AcknowledgeResponse, ErrorCode,
                                    ErrorResponse, IPAddressHoldRequest,
                                    IPAddressRequest, IPAddressResponse,
                                    Message, MessageEncDec, read_message, send_invalid_token_message,
                                    send_message)
from wgstarman.cli.server.resolver import ResolverServer
from wgstarman.cli.wgstarman_conf import WGStarManConf
from wgstarman.decorators import syncronized
from wgstarman.wg_conf import Interface, Peer, WireGuardConf
from wgstarman.wgcli.exec import WireGuardCLI

LOG_FORMAT = '%(asctime)23s %(levelname)8s %(message)s'

conf_edit_lock = Lock()


class ServerCommand(CLICommand):
    task_name = 'server'

    config: WireGuardConf
    enc_key: bytes
    end_signal: Event
    temp_address_assoc: Dict[str, Union[IPv4Address, IPv6Address]]  # TODO introduce a timeout

    ipv6_interface: IPv6Interface
    device_name: str

    @property
    def inverse_temp_address_assoc(self) -> Dict[str, str]:
        return {str(v): k for k, v in self.temp_address_assoc.items()}

    def __init__(self) -> None:
        self.temp_address_assoc = {}

    def configure_logging(self, args: Namespace) -> None:
        logging.basicConfig(format=LOG_FORMAT, level=logging.DEBUG if args.debug else logging.INFO)

    def decorate_subparsers(self, subparsers: _SubParsersAction) -> None:
        parser: ArgumentParser = subparsers.add_parser(
            self.task_name, formatter_class=ArgumentDefaultsHelpFormatter)
        required = parser.add_argument_group('required')
        optional = parser.add_argument_group('optional')

        required.add_argument('--ipv6-network', required=True, default=SUPPRESS,
                              help='the IPv6 network to use for the VPN, i.e. fdxx:xxxx:xxxx:yyyy:zzzz:zzzz::/64')
        optional.add_argument('--device-name', required=False, default='wg0',
                              help='the name of both the network device and the configuration file')
        optional.add_argument('--listen-address-4', required=False, default='0.0.0.0',
                              help='the address IPv4 on which the address manager will listen on')
        optional.add_argument('--listen-address-6', required=False, default='::',
                              help='the address IPv6 on which the address manager will listen on; will overwrite --listen-address-4 if enabled; requires --enable-listen-ipv6')
        optional.add_argument('--listen-port', required=False, default=1194, type=int,
                              help='the port on which WireGuard (UDP) and the address manager (TCP) will listen on')
        optional.add_argument('--enable-listen-ipv6', required=False, default=False, action='store_true',
                              help='enable the address manager to listen on ipv6')
        optional.add_argument('--refresh-psk', required=False, default=False, action='store_true',
                              help='force the refresh of the pre-shared key')
        optional.add_argument('--debug', required=False, default=False, action='store_true',
                              help='enable debug log')

    def can_handle(self, task_name: str) -> bool:
        return self.task_name == task_name

    def handle(self, args: Namespace) -> None:
        self.ipv6_interface = IPv6Interface(args.ipv6_network)
        self.device_name = args.device_name

        self.end_signal = Event()

        def sigint_handler(*args):
            if not self.end_signal.is_set():
                self.end_signal.set()
        signal.signal(signal.SIGINT, sigint_handler)

        # Generate an (updated) configuration
        self.config = WireGuardConf.parse(self.device_name)
        if self.config:
            interface = Interface(
                private_key=self.config.interface.private_key,
                address=self.config.interface.address,
                listen_port=args.listen_port,
                fw_mark=self.config.interface.fw_mark,
            )
            peers = self.config.peers
        else:
            interface = Interface(
                private_key=WireGuardCLI.gen_private_key(),
                address=[str(self.ipv6_interface.network[1])],
                listen_port=args.listen_port,
                fw_mark=None,
            )
            peers = []

        self.config = WireGuardConf(interface, peers)

        self.config.save(self.device_name)

        srv_conf = WGStarManConf.load()
        public_key = WireGuardCLI.get_public_key(self.config.interface.private_key)
        psk_str = srv_conf.get_preshared_key(public_key)
        if not psk_str or args.refresh_psk:
            self.enc_key = Fernet.generate_key()
            psk_str = self.enc_key.decode('utf8')
            srv_conf.set_preshared_key(public_key, psk_str)
        else:
            self.enc_key = srv_conf.get_preshared_key(public_key).encode('utf8')

        if not WireGuardCLI.up(self.device_name):
            return

        logging.info(f'WGSTARMAN PRE-SHARED KEY: {psk_str}')

        self.listen(args)

    @syncronized(conf_edit_lock)
    def _routine_ip_address(self, req: IPAddressRequest) -> Message:
        public_key = WireGuardCLI.get_public_key(self.config.interface.private_key)
        current_peer = next(
            (peer for peer in self.config.peers if peer.public_key == req.public_key), None)

        resolver_addresses = list(map(lambda x: x.split('/')[0], self.config.interface.address))

        if current_peer:
            return IPAddressResponse(
                server_public_key=public_key,
                peer_address=current_peer.allowed_ips[0],
                peer_allowed_ips=str(self.ipv6_interface.network),
                resolver_addresses=resolver_addresses,
            )

        if req.public_key in self.temp_address_assoc:
            return IPAddressResponse(
                server_public_key=public_key,
                peer_address=str(self.temp_address_assoc[public_key]),
                peer_allowed_ips=str(self.ipv6_interface.network),
                resolver_addresses=resolver_addresses,
            )

        used_ip_addresses = [ipaddress.ip_address(resolver_addresses[0])]
        used_ip_addresses.extend([ipaddress.ip_address(peer.allowed_ips[0].split('/')[0])
                                  for peer in self.config.peers])
        used_ip_addresses.extend(self.temp_address_assoc.values())

        ip_address = next(
            (
                addr for addr in self.ipv6_interface.network

                if addr not in used_ip_addresses
                and addr != self.ipv6_interface.network[0]
            ),
            None
        )

        if not ip_address:
            message = 'No address available, the network is full.'
            logging.error(message)

            return ErrorResponse(error_code=ErrorCode.NETWORK_IS_FULL.value, message=message)

        self.temp_address_assoc[req.public_key] = ip_address

        logging.info(f'Offering IP address {ip_address} to {req.public_key}')

        return IPAddressResponse(
            server_public_key=public_key,
            peer_address=str(ip_address),
            peer_allowed_ips=str(self.ipv6_interface.network),
            resolver_addresses=resolver_addresses,
        )

    @syncronized(conf_edit_lock)
    def _routine_ip_address_hold(self, req: IPAddressHoldRequest) -> Message:
        ip_address = req.ip_address

        public_key = self.inverse_temp_address_assoc[ip_address] if ip_address in self.inverse_temp_address_assoc else None
        conf_peer: Peer = next(
            (peer for peer in self.config.peers if peer.allowed_ips[0] == ip_address), None)

        if not public_key and not conf_peer:
            logging.warning(
                f'A user attempted to hold the following unreleased IP address: {ip_address}')
            return ErrorResponse(ErrorCode.INVALID_IP_ADDRESS.value, 'The requested IP has not been assigned by this server')

        public_key = public_key or conf_peer.public_key

        if req.host_name:
            peer_by_name = self.config.find_peer_by_name(req.host_name)
            if peer_by_name and peer_by_name.public_key != public_key:
                logging.warning(
                    f'A peer attempted to name itself with the name of another one: {req.host_name} - {public_key}')

                if public_key in self.temp_address_assoc:
                    del self.temp_address_assoc[public_key]

                return ErrorResponse(ErrorCode.HOST_NAME_THEFT.value, 'The requested host name has already been assigned to another peer')

        peer = Peer(public_key=public_key, allowed_ips=[
                    ip_address], name=req.host_name) if public_key else None

        if peer == conf_peer:
            return AcknowledgeResponse()

        logging.info(f'Registering IP address {ip_address} to {public_key}')

        self.config.append_peer(peer)
        self.config.save(self.device_name)

        if public_key in self.temp_address_assoc:
            del self.temp_address_assoc[public_key]

        if WireGuardCLI.hot_reload(self.device_name):
            return AcknowledgeResponse()

        logging.error('Unable to hot-reload the server configuration')

        return ErrorResponse(ErrorCode.UNABLE_TO_RELOAD_CONFIGURATION.value,
                             'Unable to update central configuration, please retry again later')

    def address_manager_routine(self, connection: socket.socket, address: Tuple[str, int]):
        try:
            raw_msg = read_message(connection, self.enc_key)

            message = MessageEncDec.loads(raw_msg)
            response: Message = None

            if message.instance_of(IPAddressRequest):
                response = self._routine_ip_address(message)
            elif message.instance_of(IPAddressHoldRequest):
                response = self._routine_ip_address_hold(message)

            if message:
                send_message(connection, response, self.enc_key)

            connection.close()
        except InvalidToken:
            send_invalid_token_message(connection)
            logging.warning(
                f'User {address[0]} attempted to connect using an invalid pre-shared key')

    def socket_listen(self, sock: socket.socket):
        threads: List[Thread] = []
        sock.settimeout(1)

        while not self.end_signal.is_set():
            try:
                sock.listen()
                connection, address = sock.accept()
                thread = Thread(target=self.address_manager_routine, args=[connection, address])
                thread.start()
                threads.append(thread)
            except socket.timeout:
                pass

        sock.close()

        # Wait for all the threads
        for thread in threads:
            if thread.is_alive():
                thread.join()

    def listen(self, args: Namespace):
        # TODO: split command server and address manager

        threads: List[Thread] = []

        if args.enable_listen_ipv6 and socket.has_dualstack_ipv6():
            sock = socket.create_server((args.listen_address_6, args.listen_port),
                                        family=socket.AF_INET6, dualstack_ipv6=True)
        else:
            sock = socket.create_server(
                (args.listen_address_4, args.listen_port), family=socket.AF_INET)

        threads.append(Thread(target=self.socket_listen, args=[sock]))

        for address in self.config.interface.address:
            resolver = ResolverServer(
                bind_address=ipaddress.ip_address(address.split('/')[0]),
                bind_port=args.listen_port + 1,
                config=self.config,
                end_signal=self.end_signal,
            )
            threads.append(Thread(target=resolver.listen))

        for thread in threads:
            thread.start()

        # Wait for the threads to end
        for thread in threads:
            thread.join()



from abc import ABC, abstractmethod
from argparse import Namespace, _SubParsersAction
from pathlib import Path
import stat

WG_CONF_FILE_MODE = stat.S_IWUSR | stat.S_IRUSR
WG_DIR_MODE = stat.S_IXUSR | stat.S_IWUSR | stat.S_IRUSR

ETC_DIR_MODE = stat.S_IWUSR | stat.S_IRUSR | stat.S_IXUSR
RESOLVERS_ETC_DIR_MODE = stat.S_IWUSR | stat.S_IRUSR | stat.S_IXUSR \
    | stat.S_IRGRP | stat.S_IXGRP \
    | stat.S_IROTH | stat.S_IXOTH
RESOLVERS_FILE_MODE = stat.S_IWUSR | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

DEFAULT_WGSTARMAN_ETC_PATH = Path('/etc/wgstarman')
DEFAULT_WGSTARMAN_CONF_PATH = str(DEFAULT_WGSTARMAN_ETC_PATH.joinpath('wgstarman.conf'))


class CLICommand(ABC):
    def configure_logging(self, args: Namespace) -> None:
        pass

    @staticmethod
    def ensure_wireguard() -> bool:
        return True

    @abstractmethod
    def decorate_subparsers(self, subparsers: _SubParsersAction) -> None:
        pass

    @abstractmethod
    def can_handle(self, task_name: str) -> bool:
        pass

    @abstractmethod
    def handle(self, args: Namespace) -> None:
        pass



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import List

from wgstarman.cli.common import CLICommand
from wgstarman.cli.peer import PeerCommand
from wgstarman.cli.resolv import ResolvCommand
from wgstarman.cli.server import ServerCommand
from wgstarman.wgcli.exec import WireGuardCLI


def main():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(dest='task')

    commands: List[CLICommand] = [PeerCommand(), ServerCommand(), ResolvCommand()]
    for command in commands:
        command.decorate_subparsers(subparsers)

    args = parser.parse_args()

    for command in commands:
        if command.can_handle(args.task):
            if command.ensure_wireguard() and not WireGuardCLI.ensure_installation():
                return

            command.configure_logging(args)
            command.handle(args)


if __name__ == '__main__':
    main()
